@@ -21,6 +21,10 @@ import bokeh.resources as _bres
 import bokeh.io as _bio
 import scipy.signal as _sig
 
+light_palette = ['mediumblue', 'crimson', 'forestgreen', 'gold', 'darkmagenta', 'olive', 'palevioletred', 'yellowgreen',
+                 'deepskyblue', 'dimgray', 'indianred', 'mediumaquamarine', 'orange', 'saddlebrown', 'teal', 'mediumorchid']
+dark_palette = ['lightskyblue', 'red', 'limegreen', 'salmon', 'magenta', 'forestgreen', 'silver', 'teal']
+
 _figure = None
 _figures = None
 _hold = False
@@ -31,8 +35,7 @@ _disable_js = False
 _using_js = False
 _interactive = True
 _static_images = False
-_colors = ['mediumblue', 'crimson', 'forestgreen', 'gold', 'darkmagenta', 'olive', 'palevioletred', 'yellowgreen',
-           'deepskyblue', 'dimgray', 'indianred', 'mediumaquamarine', 'orange', 'saddlebrown', 'teal', 'mediumorchid']
+_colors = light_palette
 
 try:
     get_ipython                     # check if we are using IPython
@@ -42,7 +45,7 @@ try:
 except:
     pass                            # not in Jupyter, skip notebook initialization
 
-def _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive):
+def _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, xtype, ytype, interactive):
     global _color
     if width is None:
         width = _figsize[0]
@@ -54,7 +57,7 @@ def _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive):
         interactive = _interactive
     if interactive:
         tools = 'pan,box_zoom,wheel_zoom,reset,save'
-    f = _bplt.figure(title=title, plot_width=width, plot_height=height, x_range=xlim, y_range=ylim, x_axis_label=xlabel, y_axis_label=ylabel, tools=tools)
+    f = _bplt.figure(title=title, plot_width=width, plot_height=height, x_range=xlim, y_range=ylim, x_axis_label=xlabel, y_axis_label=ylabel, x_axis_type=xtype, y_axis_type=ytype, tools=tools)
     f.toolbar.logo = None
     return f
 
@@ -68,7 +71,7 @@ def _process_canvas(figures):
     i = 0
     for f in figures:
         i += 1
-        if f.tools == []:
+        if f is not None and f.tools == []:
             disable.append(i)
         else:
             pass
@@ -125,6 +128,22 @@ def _hold_enable(enable):
         _figure = None
     return ohold
 
+def theme(name):
+    """Set color theme.
+
+    :param name: name of theme
+
+    >>> import arlpy.plot
+    >>> arlpy.plot.theme('dark')
+    """
+    if name == 'dark':
+        name = 'dark_minimal'
+        set_colors(dark_palette)
+    elif name == 'light':
+        name = 'light_minimal'
+        set_colors(light_palette)
+    _bio.curdoc().theme = name
+
 def figsize(x, y):
     """Set the default figure size in pixels.
 
@@ -200,6 +219,8 @@ class figure:
     :param ylabel: y-axis label
     :param xlim: x-axis limits (min, max)
     :param ylim: y-axis limits (min, max)
+    :param xtype: x-axis type ('auto', 'linear', 'log', etc)
+    :param ytype: y-axis type ('auto', 'linear', 'log', etc)
     :param width: figure width in pixels
     :param height: figure height in pixels
     :param interactive: enable interactive tools (pan, zoom, etc) for plot
@@ -225,9 +246,9 @@ class figure:
     >>>     f.square([3,7], [4,5], line_color='green', fill_color='yellow', size=10)
     """
 
-    def __init__(self, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, width=None, height=None, interactive=None):
+    def __init__(self, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, xtype='auto', ytype='auto', width=None, height=None, interactive=None):
         global _figure
-        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive)
+        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, xtype, ytype, interactive)
 
     def __enter__(self):
         global _hold
@@ -296,7 +317,7 @@ def gcf():
     """
     return _figure
 
-def plot(x, y=None, fs=None, maxpts=10000, pooling=None, color=None, style='solid', thickness=1, marker=None, filled=False, size=6, mskip=0, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, width=None, height=None, legend=None, hold=False, interactive=None):
+def plot(x, y=None, fs=None, maxpts=10000, pooling=None, color=None, style='solid', thickness=1, marker=None, filled=False, size=6, mskip=0, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, xtype='auto', ytype='auto', width=None, height=None, legend=None, hold=False, interactive=None):
     """Plot a line graph or time series.
 
     :param x: x data or time series data (if y is None)
@@ -316,6 +337,8 @@ def plot(x, y=None, fs=None, maxpts=10000, pooling=None, color=None, style='soli
     :param ylabel: y-axis label
     :param xlim: x-axis limits (min, max)
     :param ylim: y-axis limits (min, max)
+    :param xtype: x-axis type ('auto', 'linear', 'log', etc)
+    :param ytype: y-axis type ('auto', 'linear', 'log', etc)
     :param width: figure width in pixels
     :param height: figure height in pixels
     :param legend: legend text
@@ -341,7 +364,7 @@ def plot(x, y=None, fs=None, maxpts=10000, pooling=None, color=None, style='soli
     else:
         y = _np.array(y, ndmin=1, dtype=_np.float, copy=False)
     if _figure is None:
-        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive)
+        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, xtype, ytype, interactive)
     if color is None:
         color = _colors[_color % len(_colors)]
         _color += 1
@@ -366,16 +389,21 @@ def plot(x, y=None, fs=None, maxpts=10000, pooling=None, color=None, style='soli
         else:
             _warnings.warn('Unknown pooling: '+pooling)
             y = y[::n]
+        if len(x) > len(y):
+            x = x[:len(y)]
         _figure.add_layout(_bmodels.Label(x=5, y=5, x_units='screen', y_units='screen', text=desc, text_font_size="8pt", text_alpha=0.5))
     if style is not None:
-        _figure.line(x, y, line_color=color, line_dash=style, line_width=thickness, legend=legend)
+        if legend is None:
+            _figure.line(x, y, line_color=color, line_dash=style, line_width=thickness)
+        else:
+            _figure.line(x, y, line_color=color, line_dash=style, line_width=thickness, legend_label=legend)
     if marker is not None:
         scatter(x[::(mskip+1)], y[::(mskip+1)], marker=marker, filled=filled, size=size, color=color, legend=legend, hold=True)
     if not hold and not _hold:
         _show(_figure)
         _figure = None
 
-def scatter(x, y, marker='.', filled=False, size=6, color=None, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, width=None, height=None, legend=None, hold=False, interactive=None):
+def scatter(x, y, marker='.', filled=False, size=6, color=None, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, xtype='auto', ytype='auto', width=None, height=None, legend=None, hold=False, interactive=None):
     """Plot a scatter plot.
 
     :param x: x data
@@ -389,6 +417,8 @@ def scatter(x, y, marker='.', filled=False, size=6, color=None, title=None, xlab
     :param ylabel: y-axis label
     :param xlim: x-axis limits (min, max)
     :param ylim: y-axis limits (min, max)
+    :param xtype: x-axis type ('auto', 'linear', 'log', etc)
+    :param ytype: y-axis type ('auto', 'linear', 'log', etc)
     :param width: figure width in pixels
     :param height: figure height in pixels
     :param legend: legend text
@@ -401,35 +431,42 @@ def scatter(x, y, marker='.', filled=False, size=6, color=None, title=None, xlab
     """
     global _figure, _color
     if _figure is None:
-        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive)
+        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, xtype, ytype, interactive)
     x = _np.array(x, ndmin=1, dtype=_np.float, copy=False)
     y = _np.array(y, ndmin=1, dtype=_np.float, copy=False)
     if color is None:
         color = _colors[_color % len(_colors)]
         _color += 1
+    kwargs = {'size': size, 'line_color': color}
+    if filled:
+        kwargs['fill_color'] = color
+    if legend is not None:
+        kwargs['legend_label'] = legend
     if marker == '.':
-        _figure.circle(x, y, size=size/2, line_color=color, fill_color=color, legend=legend)
+        kwargs['size'] = kwargs['size']/2
+        kwargs['fill_color'] = color
+        _figure.circle(x, y, **kwargs)
     elif marker == 'o':
-        _figure.circle(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.circle(x, y, **kwargs)
     elif marker == 's':
-        _figure.square(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.square(x, y, **kwargs)
     elif marker == '*':
-        _figure.asterisk(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.asterisk(x, y, **kwargs)
     elif marker == 'x':
-        _figure.x(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.x(x, y, **kwargs)
     elif marker == '+':
-        _figure.cross(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.cross(x, y, **kwargs)
     elif marker == 'd':
-        _figure.diamond(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.diamond(x, y, **kwargs)
     elif marker == '^':
-        _figure.triangle(x, y, size=size, line_color=color, fill_color=color if filled else None, legend=legend)
+        _figure.triangle(x, y, **kwargs)
     elif marker is not None:
         _warnings.warn('Bad marker type: '+marker)
     if not hold and not _hold:
         _show(_figure)
         _figure = None
 
-def image(img, x=None, y=None, colormap='Plasma256', clim=None, clabel=None, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, width=None, height=None, hold=False, interactive=None):
+def image(img, x=None, y=None, colormap='Plasma256', clim=None, clabel=None, title=None, xlabel=None, ylabel=None, xlim=None, ylim=None, xtype='auto', ytype='auto', width=None, height=None, hold=False, interactive=None):
     """Plot a heatmap of 2D scalar data.
 
     :param img: 2D image data
@@ -443,6 +480,8 @@ def image(img, x=None, y=None, colormap='Plasma256', clim=None, clabel=None, tit
     :param ylabel: y-axis label
     :param xlim: x-axis limits (min, max)
     :param ylim: y-axis limits (min, max)
+    :param xtype: x-axis type ('auto', 'linear', 'log', etc)
+    :param ytype: y-axis type ('auto', 'linear', 'log', etc)
     :param width: figure width in pixels
     :param height: figure height in pixels
     :param interactive: enable interactive tools (pan, zoom, etc) for plot
@@ -462,10 +501,10 @@ def image(img, x=None, y=None, colormap='Plasma256', clim=None, clabel=None, tit
     if ylim is None:
         ylim = y
     if _figure is None:
-        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, interactive)
+        _figure = _new_figure(title, width, height, xlabel, ylabel, xlim, ylim, xtype, ytype, interactive)
     if clim is None:
         clim = [_np.amin(img), _np.amax(img)]
-    if isinstance(colormap, str):
+    if not isinstance(colormap, _bmodels.ColorMapper):
         colormap = _bmodels.LinearColorMapper(palette=colormap, low=clim[0], high=clim[1])
     _figure.image([img], x=x[0], y=y[0], dw=x[-1]-x[0], dh=y[-1]-y[0], color_mapper=colormap)
     cbar = _bmodels.ColorBar(color_mapper=colormap, location=(0,0), title=clabel)
@@ -612,13 +651,13 @@ def specgram(x, fs=2, nfft=None, noverlap=None, colormap='Plasma256', clim=None,
     >>> import numpy as np
     >>> arlpy.plot.specgram(np.random.normal(size=(10000)), fs=10000, clim=30)
     """
-    f, t, Sxx = _sig.spectrogram(x, fs=fs, nfft=nfft, noverlap=noverlap)
-    Sxx = 10*_np.log10(Sxx)
+    f, t, Sxx = _sig.spectrogram(x, fs=fs, nperseg=nfft, noverlap=noverlap)
+    Sxx = 10*_np.log10(Sxx+_np.finfo(float).eps)
     if isinstance(clim, float) or isinstance(clim, int):
         clim = (_np.max(Sxx)-clim, _np.max(Sxx))
     image(Sxx, x=(t[0], t[-1]), y=(f[0], f[-1]), title=title, colormap=colormap, clim=clim, clabel=clabel, xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, width=width, height=height, hold=hold, interactive=interactive)
 
-def psd(x, fs=2, nfft=512, noverlap=None, window='hanning', color=None, style='solid', thickness=1, marker=None, filled=False, size=6, title=None, xlabel='Frequency (Hz)', ylabel='Power spectral density (dB/Hz)', xlim=None, ylim=None, width=None, height=None, hold=False, interactive=None):
+def psd(x, fs=2, nfft=512, noverlap=None, window='hanning', color=None, style='solid', thickness=1, marker=None, filled=False, size=6, title=None, xlabel='Frequency (Hz)', ylabel='Power spectral density (dB/Hz)', xlim=None, ylim=None, width=None, height=None, legend=None, hold=False, interactive=None):
     """Plot power spectral density of a given time series signal.
 
     :param x: time series signal
@@ -639,6 +678,7 @@ def psd(x, fs=2, nfft=512, noverlap=None, window='hanning', color=None, style='s
     :param ylim: y-axis limits (min, max)
     :param width: figure width in pixels
     :param height: figure height in pixels
+    :param legend: legend text
     :param interactive: enable interactive tools (pan, zoom, etc) for plot
     :param hold: if set to True, output is not plotted immediately, but combined with the next plot
 
@@ -647,12 +687,12 @@ def psd(x, fs=2, nfft=512, noverlap=None, window='hanning', color=None, style='s
     >>> arlpy.plot.psd(np.random.normal(size=(10000)), fs=10000)
     """
     f, Pxx = _sig.welch(x, fs=fs, nperseg=nfft, noverlap=noverlap, window=window)
-    Pxx = 10*_np.log10(Pxx)
+    Pxx = 10*_np.log10(Pxx+_np.finfo(float).eps)
     if xlim is None:
         xlim = (0, fs/2)
     if ylim is None:
         ylim = (_np.max(Pxx)-50, _np.max(Pxx)+10)
-    plot(f, Pxx, color=color, style=style, thickness=thickness, marker=marker, filled=filled, size=size, title=title, xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, width=width, height=height, hold=hold, interactive=interactive)
+    plot(f, Pxx, color=color, style=style, thickness=thickness, marker=marker, filled=filled, size=size, title=title, xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, maxpts=len(f), width=width, height=height, hold=hold, legend=legend, interactive=interactive)
 
 def iqplot(data, marker='.', color=None, labels=None, filled=False, size=None, title=None, xlabel=None, ylabel=None, xlim=[-2, 2], ylim=[-2, 2], width=None, height=None, hold=False, interactive=None):
     """Plot signal points.
@@ -680,7 +720,8 @@ def iqplot(data, marker='.', color=None, labels=None, filled=False, size=None, t
     >>> arlpy.plot.iqplot(arlpy.comms.psk(4), labels=['00', '01', '11', '10'])
     """
     data = _np.asarray(data, dtype=_np.complex)
-    figure(title=title, xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, width=width, height=height, interactive=interactive)
+    if not _hold:
+        figure(title=title, xlabel=xlabel, ylabel=ylabel, xlim=xlim, ylim=ylim, width=width, height=height, interactive=interactive)
     if labels is None:
         if size is None:
             size = 5
@@ -738,5 +779,5 @@ def freqz(b, a=1, fs=2.0, worN=None, whole=False, degrees=True, style='solid', t
     fig.extra_y_ranges = {'phase': _bmodels.Range1d(start=-_np.pi*units, end=_np.pi*units)}
     fig.add_layout(_bmodels.LinearAxis(y_range_name='phase', axis_label='Phase (degrees)' if degrees else 'Phase (radians)'), 'right')
     phase = _np.angle(h)*units
-    fig.line(f, phase, line_color=color(1), line_dash=style, line_width=thickness, legend='Phase', y_range_name='phase')
+    fig.line(f, phase, line_color=color(1), line_dash=style, line_width=thickness, legend_label='Phase', y_range_name='phase')
     _hold_enable(hold)
