@@ -11,7 +11,7 @@
 """Underwater acoustic propagation modeling toolbox.
 
 This toolbox currently uses the Bellhop acoustic propagation model. For this model
-to work, the `acoustic toolbox <http://oalib.hlsresearch.com/Modes/AcousticsToolbox/>`_
+to work, the `acoustic toolbox <https://oalib-acoustics.org/>`_
 must be installed on your computer and `bellhop.exe` should be in your PATH.
 
 .. sidebar:: Sample Jupyter notebook
@@ -69,7 +69,18 @@ def create_env2d(**kv):
     profile be provided as a Nx2 array of (depth, sound speed):
 
     >>> import arlpy.uwapm as pm
-    >>> env = pm.create_env2d(depth=20, soundspeed=[[0,1540], [5,1535], [20,1530]])
+    >>> env = pm.create_env2d(depth=20, soundspeed=[[0,1540], [5,1535], [10,1535], [20,1530]])
+
+    A range-and-depth dependent sound speed profile can be provided as a Pandas frame:
+
+    >>> import arlpy.uwapm as pm
+    >>> import pandas as pd
+    >>> ssp2 = pd.DataFrame({
+              0: [1540, 1530, 1532, 1533],     # profile at 0 m range
+            100: [1540, 1535, 1530, 1533],     # profile at 100 m range
+            200: [1530, 1520, 1522, 1525] },   # profile at 200 m range
+            index=[0, 10, 20, 30])             # depths of the profile entries in m
+    >>> env = pm.create_env2d(depth=20, soundspeed=ssp2)
 
     The default environment has a constant water depth. A range dependent bathymetry
     can be provided as a Nx2 array of (range, water depth):
@@ -96,13 +107,14 @@ def create_env2d(**kv):
         'depth': 25,                    # m
         'depth_interp': linear,         # curvilinear/linear
         'min_angle': -80,               # deg
-        'max_angle': 80                 # deg
+        'max_angle': 80,                # deg
+        'nbeams': 0                     # number of beams (0 = auto)
     }
     for k, v in kv.items():
         if k not in env.keys():
             raise KeyError('Unknown key: '+k)
-        env[k] = _np.asarray(v, dtype=_np.float) if _np.size(v) > 1 else v
-    check_env2d(env)
+        env[k] = _np.asarray(v, dtype=_np.float) if not isinstance(v, _pd.DataFrame) and _np.size(v) > 1 else v
+    env = check_env2d(env)
     return env
 
 def check_env2d(env):
@@ -137,7 +149,12 @@ def check_env2d(env):
             max_depth = _np.max(env['depth'][:,1])
         else:
             max_depth = env['depth']
-        if _np.size(env['soundspeed']) > 1:
+        if isinstance(env['soundspeed'], _pd.DataFrame):
+            assert env['soundspeed'].shape[0] > 3, 'soundspeed profile must have at least 4 points'
+            assert env['soundspeed'].index[0] <= 0, 'First depth in soundspeed array must be 0 m'
+            assert env['soundspeed'].index[-1] >= max_depth, 'Last depth in soundspeed array must be beyond water depth: '+str(max_depth)+' m'
+            assert _np.all(_np.diff(env['soundspeed'].index) > 0), 'Soundspeed array must be strictly monotonic in depth'
+        elif _np.size(env['soundspeed']) > 1:
             assert env['soundspeed'].ndim == 2, 'soundspeed must be a scalar or an Nx2 array'
             assert env['soundspeed'].shape[1] == 2, 'soundspeed must be a scalar or an Nx2 array'
             assert env['soundspeed'].shape[0] > 3, 'soundspeed profile must have at least 4 points'
@@ -145,6 +162,15 @@ def check_env2d(env):
             assert env['soundspeed'][-1,0] >= max_depth, 'Last depth in soundspeed array must be beyond water depth: '+str(max_depth)+' m'
             assert _np.all(_np.diff(env['soundspeed'][:,0]) > 0), 'Soundspeed array must be strictly monotonic in depth'
             assert env['soundspeed_interp'] == spline or env['soundspeed_interp'] == linear, 'Invalid interpolation type: '+str(env['soundspeed_interp'])
+            if not(max_depth in env['soundspeed'][:,0]):
+                indlarger = _np.argwhere(env['soundspeed'][:,0]>max_depth)[0][0]
+                if env['soundspeed_interp'] == spline:
+                    tck = _interp.splrep(env['soundspeed'][:,0], env['soundspeed'][:,1], s=0)
+                    insert_ss_val = _interp.splev(max_depth, tck, der=0)
+                else:
+                    insert_ss_val = _np.interp(max_depth, env['soundspeed'][:,0], env['soundspeed'][:,1])
+                env['soundspeed'] = _np.insert(env['soundspeed'],indlarger,[max_depth,insert_ss_val],axis = 0)
+                env['soundspeed'] = env['soundspeed'][:indlarger+1,:]
         assert _np.max(env['tx_depth']) <= max_depth, 'tx_depth cannot exceed water depth: '+str(max_depth)+' m'
         assert _np.max(env['rx_depth']) <= max_depth, 'rx_depth cannot exceed water depth: '+str(max_depth)+' m'
         assert env['min_angle'] > -90 and env['min_angle'] < 90, 'min_angle must be in range (-90, 90)'
@@ -154,6 +180,7 @@ def check_env2d(env):
             assert env['tx_directionality'].ndim == 2, 'tx_directionality must be an Nx2 array'
             assert env['tx_directionality'].shape[1] == 2, 'tx_directionality must be an Nx2 array'
             assert _np.all(env['tx_directionality'][:,0] >= -180) and _np.all(env['tx_directionality'][:,0] <= 180), 'tx_directionality angles must be in [-90, 90]'
+        return env
     except AssertionError as e:
         raise ValueError(e.args)
 
@@ -166,7 +193,7 @@ def print_env(env):
     >>> env = pm.create_env2d(depth=40, soundspeed=1540)
     >>> pm.print_env(env)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     keys = ['name'] + sorted(list(env.keys()-['name']))
     for k in keys:
         v = str(env[k])
@@ -198,7 +225,7 @@ def plot_env(env, surface_color='dodgerblue', bottom_color='peru', tx_color='ora
     >>> env = pm.create_env2d(depth=[[0, 40], [100, 30], [500, 35], [700, 20], [1000,45]])
     >>> pm.plot_env(env)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     min_x = 0
     max_x = _np.max(env['rx_range'])
     if max_x-min_x > 10000:
@@ -252,27 +279,31 @@ def plot_ssp(env, **kwargs):
 
     Other keyword arguments applicable for `arlpy.plot.plot()` are also supported.
 
+    If the sound speed profile is range-dependent, this function only plots the first profile.
+
     >>> import arlpy.uwapm as pm
     >>> env = pm.create_env2d(soundspeed=[[ 0, 1540], [10, 1530], [20, 1532], [25, 1533], [30, 1535]])
     >>> pm.plot_ssp(env)
     """
-    check_env2d(env)
-    if _np.size(env['soundspeed']) == 1:
+    env = check_env2d(env)
+    svp = env['soundspeed']
+    if isinstance(svp, _pd.DataFrame):
+        svp = _np.hstack((_np.array([svp.index]).T, _np.asarray(svp)))
+    if _np.size(svp) == 1:
         if _np.size(env['depth']) > 1:
             max_y = _np.max(env['depth'][:,1])
         else:
             max_y = env['depth']
-        _plt.plot([env['soundspeed'], env['soundspeed']], [0, -max_y], xlabel='Soundspeed (m/s)', ylabel='Depth (m)', **kwargs)
+        _plt.plot([svp, svp], [0, -max_y], xlabel='Soundspeed (m/s)', ylabel='Depth (m)', **kwargs)
     elif env['soundspeed_interp'] == spline:
-        s = env['soundspeed']
-        ynew = _np.linspace(_np.min(s[:,0]), _np.max(s[:,0]), 100)
-        tck = _interp.splrep(s[:,0], s[:,1], s=0)
+        s = svp
+        ynew = _np.linspace(_np.min(svp[:,0]), _np.max(svp[:,0]), 100)
+        tck = _interp.splrep(svp[:,0], svp[:,1], s=0)
         xnew = _interp.splev(ynew, tck, der=0)
         _plt.plot(xnew, -ynew, xlabel='Soundspeed (m/s)', ylabel='Depth (m)', hold=True, **kwargs)
-        _plt.plot(s[:,1], -s[:,0], marker='.', style=None, **kwargs)
+        _plt.plot(svp[:,1], -svp[:,0], marker='.', style=None, **kwargs)
     else:
-        s = env['soundspeed']
-        _plt.plot(s[:,1], -s[:,0], xlabel='Soundspeed (m/s)', ylabel='Depth (m)', **kwargs)
+        _plt.plot(svp[:,1], -svp[:,0], xlabel='Soundspeed (m/s)', ylabel='Depth (m)', **kwargs)
 
 def compute_arrivals(env, model=None, debug=False):
     """Compute arrivals between each transmitter and receiver.
@@ -287,7 +318,7 @@ def compute_arrivals(env, model=None, debug=False):
     >>> arrivals = pm.compute_arrivals(env)
     >>> pm.plot_arrivals(arrivals)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     (model_name, model) = _select_model(env, arrivals, model)
     if debug:
         print('[DEBUG] Model: '+model_name)
@@ -309,7 +340,7 @@ def compute_eigenrays(env, tx_depth_ndx=0, rx_depth_ndx=0, rx_range_ndx=0, model
     >>> rays = pm.compute_eigenrays(env)
     >>> pm.plot_rays(rays, width=1000)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     env = env.copy()
     if _np.size(env['tx_depth']) > 1:
         env['tx_depth'] = env['tx_depth'][tx_depth_ndx]
@@ -336,7 +367,7 @@ def compute_rays(env, tx_depth_ndx=0, model=None, debug=False):
     >>> rays = pm.compute_rays(env)
     >>> pm.plot_rays(rays, width=1000)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     if _np.size(env['tx_depth']) > 1:
         env = env.copy()
         env['tx_depth'] = env['tx_depth'][tx_depth_ndx]
@@ -360,7 +391,7 @@ def compute_transmission_loss(env, tx_depth_ndx=0, mode=coherent, model=None, de
     >>> tloss = pm.compute_transmission_loss(env, mode=pm.incoherent)
     >>> pm.plot_transmission_loss(tloss, width=1000)
     """
-    check_env2d(env)
+    env = check_env2d(env)
     if mode not in [coherent, incoherent, semicoherent]:
         raise ValueError('Unknown transmission loss mode: '+mode)
     if _np.size(env['tx_depth']) > 1:
@@ -427,11 +458,12 @@ def plot_arrivals(arrivals, dB=False, color='blue', **kwargs):
         _plt.plot([t, t], [min_y, y], xlabel='Arrival time (s)', ylabel=ylabel, ylim=[min_y, min_y+70], color=color, **kwargs)
     _plt.hold(oh)
 
-def plot_rays(rays, env=None, **kwargs):
+def plot_rays(rays, env=None, invert_colors=False, **kwargs):
     """Plots ray paths.
 
     :param rays: ray paths
     :param env: environment definition
+    :param invert_colors: False to use black for high intensity rays, True to use white
 
     If environment definition is provided, it is overlayed over this plot using default
     parameters for `arlpy.uwapm.plot_env()`.
@@ -444,7 +476,9 @@ def plot_rays(rays, env=None, **kwargs):
     >>> pm.plot_rays(rays, width=1000)
     """
     rays = rays.sort_values('bottom_bounces', ascending=False)
-    max_amp = _np.max(_np.abs(rays.bottom_bounces))
+    max_amp = _np.max(_np.abs(rays.bottom_bounces)) if len(rays.bottom_bounces) > 0 else 0
+    if max_amp <= 0:
+        max_amp = 1
     divisor = 1
     xlabel = 'Range (m)'
     r = []
@@ -456,6 +490,8 @@ def plot_rays(rays, env=None, **kwargs):
     oh = _plt.hold()
     for _, row in rays.iterrows():
         c = int(255*_np.abs(row.bottom_bounces)/max_amp)
+        if invert_colors:
+            c = 255-c
         c = _bokeh.colors.RGB(c, c, c)
         _plt.plot(row.ray[:,0]/divisor, -row.ray[:,1], color=c, xlabel=xlabel, ylabel='Depth (m)', **kwargs)
     if env is not None:
@@ -511,7 +547,7 @@ def models(env=None, task=None):
     ['bellhop']
     """
     if env is not None:
-        check_env2d(env)
+        env = check_env2d(env)
     if (env is None and task is not None) or (env is not None and task is None):
         raise ValueError('env and task should be both specified together')
     rv = []
@@ -545,8 +581,8 @@ class _Bellhop:
         fh, fname = _mkstemp(suffix='.env')
         _os.close(fh)
         fname_base = fname[:-4]
-        rv = self._bellhop(fname_base)
         self._unlink(fname_base+'.env')
+        rv = self._bellhop(fname_base)
         self._unlink(fname_base+'.prt')
         self._unlink(fname_base+'.log')
         return rv
@@ -561,15 +597,22 @@ class _Bellhop:
             semicoherent: ['S', self._load_shd]
         }
         fname_base = self._create_env_file(env, taskmap[task][0])
+        results = None
         if self._bellhop(fname_base):
-            results = taskmap[task][1](fname_base)
-        else:
-            results = None
+            err = self._check_error(fname_base)
+            if err is not None:
+                print(err)
+            else:
+                try:
+                    results = taskmap[task][1](fname_base)
+                except FileNotFoundError:
+                    print('[WARN] Bellhop did not generate expected output file')
         if debug:
             print('[DEBUG] Bellhop working files: '+fname_base+'.*')
         else:
             self._unlink(fname_base+'.env')
             self._unlink(fname_base+'.bty')
+            self._unlink(fname_base+'.ssp')
             self._unlink(fname_base+'.ati')
             self._unlink(fname_base+'.sbp')
             self._unlink(fname_base+'.prt')
@@ -588,7 +631,7 @@ class _Bellhop:
 
     def _unlink(self, f):
         try:
-            os.unlink(f)
+            _os.unlink(f)
         except:
             pass
 
@@ -598,40 +641,50 @@ class _Bellhop:
     def _print_array(self, fh, a):
         if _np.size(a) == 1:
             self._print(fh, "1")
-            self._print(fh, "%0.4f /" % (a))
+            self._print(fh, "%0.6f /" % (a))
         else:
             self._print(fh, str(_np.size(a)))
             for j in a:
-                self._print(fh, "%0.4f " % (j), newline=False)
+                self._print(fh, "%0.6f " % (j), newline=False)
             self._print(fh, "/")
 
     def _create_env_file(self, env, taskcode):
         fh, fname = _mkstemp(suffix='.env')
         fname_base = fname[:-4]
         self._print(fh, "'"+env['name']+"'")
-        self._print(fh, "%0.4f" % (env['frequency']))
+        self._print(fh, "%0.6f" % (env['frequency']))
         self._print(fh, "1")
+        svp = env['soundspeed']
+        svp_interp = 'S' if env['soundspeed_interp'] == spline else 'C'
+        if isinstance(svp, _pd.DataFrame):
+            if len(svp.columns) > 1:
+                svp_interp = 'Q'
+            else:
+                svp = _np.hstack((_np.array([svp.index]).T, _np.asarray(svp)))
         if env['surface'] is None:
-            self._print(fh, "'%cVWT'" % ('S' if env['soundspeed_interp'] == spline else 'C'))
+            self._print(fh, "'%cVWT'" % svp_interp)
         else:
-            self._print(fh, "'%cVWT*'" % ('S' if env['soundspeed_interp'] == spline else 'C'))
+            self._print(fh, "'%cVWT*'" % svp_interp)
             self._create_bty_ati_file(fname_base+'.ati', env['surface'], env['surface_interp'])
         max_depth = env['depth'] if _np.size(env['depth']) == 1 else _np.max(env['depth'][:,1])
-        self._print(fh, "1 0.0 %0.4f" % (max_depth))
-        svp = env['soundspeed']
+        self._print(fh, "1 0.0 %0.6f" % (max_depth))
         if _np.size(svp) == 1:
-            self._print(fh, "0.0 %0.4f /" % (svp))
-            self._print(fh, "%0.4f %0.4f /" % (max_depth, svp))
+            self._print(fh, "0.0 %0.6f /" % (svp))
+            self._print(fh, "%0.6f %0.6f /" % (max_depth, svp))
+        elif svp_interp == 'Q':
+            for j in range(svp.shape[0]):
+                self._print(fh, "%0.6f %0.6f /" % (svp.index[j], svp.iloc[j,0]))
+            self._create_ssp_file(fname_base+'.ssp', svp)
         else:
             for j in range(svp.shape[0]):
-                self._print(fh, "%0.4f %0.4f /" % (svp[j,0], svp[j,1]))
+                self._print(fh, "%0.6f %0.6f /" % (svp[j,0], svp[j,1]))
         depth = env['depth']
         if _np.size(depth) == 1:
-            self._print(fh, "'A' %0.4f" % (env['bottom_roughness']))
+            self._print(fh, "'A' %0.6f" % (env['bottom_roughness']))
         else:
-            self._print(fh, "'A*' %0.4f" % (env['bottom_roughness']))
+            self._print(fh, "'A*' %0.6f" % (env['bottom_roughness']))
             self._create_bty_ati_file(fname_base+'.bty', depth, env['depth_interp'])
-        self._print(fh, "%0.4f %0.4f 0.0 %0.4f %0.4f /" % (max_depth, env['bottom_soundspeed'], env['bottom_density']/1000, env['bottom_absorption']))
+        self._print(fh, "%0.6f %0.6f 0.0 %0.6f %0.6f /" % (max_depth, env['bottom_soundspeed'], env['bottom_density']/1000, env['bottom_absorption']))
         self._print_array(fh, env['tx_depth'])
         self._print_array(fh, env['rx_depth'])
         self._print_array(fh, env['rx_range']/1000)
@@ -640,9 +693,9 @@ class _Bellhop:
         else:
             self._print(fh, "'"+taskcode+" *'")
             self._create_sbp_file(fname_base+'.sbp', env['tx_directionality'])
-        self._print(fh, "0")
-        self._print(fh, "%0.4f %0.4f /" % (env['min_angle'], env['max_angle']))
-        self._print(fh, "0.0 %0.4f %0.4f" % (1.01*max_depth, 1.01*_np.max(env['rx_range'])/1000))
+        self._print(fh, "%d" % (env['nbeams']))
+        self._print(fh, "%0.6f %0.6f /" % (env['min_angle'], env['max_angle']))
+        self._print(fh, "0.0 %0.6f %0.6f" % (1.01*max_depth, 1.01*_np.max(env['rx_range'])/1000))
         _os.close(fh)
         return fname_base
 
@@ -651,27 +704,70 @@ class _Bellhop:
             f.write("'%c'\n" % ('C' if interp == curvilinear else 'L'))
             f.write(str(depth.shape[0])+"\n")
             for j in range(depth.shape[0]):
-                f.write("%0.4f %0.4f\n" % (depth[j,0]/1000, depth[j,1]))
+                f.write("%0.6f %0.6f\n" % (depth[j,0]/1000, depth[j,1]))
 
     def _create_sbp_file(self, filename, dir):
         with open(filename, 'wt') as f:
             f.write(str(dir.shape[0])+"\n")
             for j in range(dir.shape[0]):
-                f.write("%0.4f %0.4f\n" % (dir[j,0], dir[j,1]))
+                f.write("%0.6f %0.6f\n" % (dir[j,0], dir[j,1]))
 
-    def _readf(self, f, types):
-        p = _re.split(r' +', f.readline().strip())
+    def _create_ssp_file(self, filename, svp):
+        with open(filename, 'wt') as f:
+            f.write(str(svp.shape[1])+"\n")
+            for j in range(svp.shape[1]):
+                f.write("%0.6f%c" % (svp.columns[j]/1000, '\n' if j == svp.shape[1]-1 else ' '))
+            for k in range(svp.shape[0]):
+                for j in range(svp.shape[1]):
+                    f.write("%0.6f%c" % (svp.iloc[k,j], '\n' if j == svp.shape[1]-1 else ' '))
+
+    def _readf(self, f, types, dtype=str):
+        if type(f) is str:
+            p = _re.split(r' +', f.strip())
+        else:
+            p = _re.split(r' +', f.readline().strip())
         for j in range(len(p)):
             if len(types) > j:
                 p[j] = types[j](p[j])
+            else:
+                p[j] = dtype(p[j])
         return tuple(p)
 
+    def _check_error(self, fname_base):
+        err = None
+        try:
+            with open(fname_base+'.prt', 'rt') as f:
+                for lno, s in enumerate(f):
+                    if err is not None:
+                        err += '[BELLHOP] ' + s
+                    elif '*** FATAL ERROR ***' in s:
+                        err = '\n[BELLHOP] ' + s
+        except:
+            pass
+        return err
+
     def _load_arrivals(self, fname_base):
         with open(fname_base+'.arr', 'rt') as f:
-            freq, tx_depth_count, rx_depth_count, rx_range_count = self._readf(f, (float, int, int, int))
-            tx_depth = self._readf(f, (float,)*tx_depth_count)
-            rx_depth = self._readf(f, (float,)*rx_depth_count)
-            rx_range = self._readf(f, (float,)*rx_range_count)
+            hdr = f.readline()
+            if hdr.find('2D') >= 0:
+                freq = self._readf(f, (float,))
+                tx_depth_info = self._readf(f, (int,), float)
+                tx_depth_count = tx_depth_info[0]
+                tx_depth = tx_depth_info[1:]
+                assert tx_depth_count == len(tx_depth)
+                rx_depth_info = self._readf(f, (int,), float)
+                rx_depth_count = rx_depth_info[0]
+                rx_depth = rx_depth_info[1:]
+                assert rx_depth_count == len(rx_depth)
+                rx_range_info = self._readf(f, (int,), float)
+                rx_range_count = rx_range_info[0]
+                rx_range = rx_range_info[1:]
+                assert rx_range_count == len(rx_range)
+            else:
+                freq, tx_depth_count, rx_depth_count, rx_range_count = self._readf(hdr, (float, int, int, int))
+                tx_depth = self._readf(f, (float,)*tx_depth_count)
+                rx_depth = self._readf(f, (float,)*rx_depth_count)
+                rx_range = self._readf(f, (float,)*rx_range_count)
             arrivals = []
             for j in range(tx_depth_count):
                 f.readline()
